from contextlib import contextmanager
from pathlib import Path
import time
import uuid

import numpy as np
import pickle
from scipy.optimize import curve_fit


class Buffer:
    def __init__(self, nbuff, target_directory):
        self.nbuff = nbuff
        self.counter = 0
        self.queue = []
        self.target_directory = Path(target_directory)

    def flush(self):
        if self.counter > 0:
            path = self.target_directory / Path(f"{uuid.uuid4().hex}.pkl")
            pickle.dump(self.queue, open(path, "wb"), protocol=4)
            self.counter = 0
            self.queue = []

    def __call__(self, val):
        self.queue.append(val)
        self.counter += 1
        if self.counter >= self.nbuff:
            self.flush()


def chunk_jobs(jobs, world_size, rank):
    return np.array_split(jobs, world_size)[rank].tolist()


def padded_kw(k, w, num_brig, ext=20):
    """For two arrays of given width and chunk size,
    this gives the optimal padding amount. ext
    fixes how far are we willing to go."""

    # if w is not 1, pad w, otherwise pad k
    if len(w) > 1:
        rems = (
            len(k) * np.array([len(w) + ii for ii in range(ext)])
        ) % num_brig
        num_pads = np.where(rems == 0)[0][0]
        w_new = np.zeros(len(w) + num_pads)
        dw = w[-1] - w[-2]
        pad_arr = np.array([w[-1] + ii * dw for ii in range(1, num_pads + 1)])
        w_new[: len(w)] = w
        w_new[len(w) :] = pad_arr
        return k, w_new
    else:
        rems = (
            np.array([len(k) + ii for ii in range(ext)]) * len(w)
        ) % num_brig
        num_pads = np.where(rems == 0)[0][0]
        k_new = np.zeros(len(k) + num_pads)
        dk = k[-1] - k[-2]
        pad_arr = np.array([k[-1] + ii * dk for ii in range(1, num_pads + 1)])
        k_new[: len(k)] = k
        k_new[len(k) :] = pad_arr
        return k_new, w


def float_to_list(val):
    if isinstance(val, float):
        val = [val]
    return val


def flatten(t):
    return [item for sublist in t for item in sublist]


def time_remaining(time_elapsed, percentage_complete):
    """Returns the time remaining."""

    # time_elapsed / percent_elapsed = time_remaining / pc_remaining
    # time_remaining = time_elapased / percent_elapsed * pc_remaining
    if percentage_complete == 100:
        return 0.0
    return (100.0 - percentage_complete) * time_elapsed / percentage_complete


def _elapsed_time_str(dt):
    """Returns the elapsed time in variable format depending on how long
    the calculation took.
    Parameters
    ----------
    dt : {float}
        The elapsed time in seconds.
    Returns
    -------
    float, str
        The elapsed time in the format given by the second returned value.
        Either seconds, minutes, hours or days.
    """

    if dt < 10.0:
        return dt, "s"
    elif 10.0 <= dt < 600.0:  # 10 s <= dt < 10 m
        return dt / 60.0, "m"
    elif 600.0 <= dt < 36000.0:  # 10 m <= dt < 10 h
        return dt / 3600.0, "h"
    else:
        return dt / 86400.0, "d"


def _adjust_log_msg_for_time(msg, elapsed):
    if elapsed is None:
        return msg
    (elapsed, units) = _elapsed_time_str(elapsed)
    return f"[{elapsed:.02f} {units}] {msg}"


@contextmanager
def timeit(stream, msg):
    """A simple utility for timing how long a certain block of code will take.

    .. hint::

        Here's an example::

            with timeit(logger.info, "Test foo"):
                foo()

    Parameters
    ----------
    stream : callable
        The callable function/method which must take only a string as an
        argument. Usually something like ``logger.info``.
    msg : str
        The message to pass to the ``stream``.
    """

    t0 = time.time()
    try:
        yield None
    finally:
        dt = time.time() - t0
        stream(_adjust_log_msg_for_time(msg, dt))


def peak_location_and_weight(w, A, Aprime, eta, eta_prime):
    """Assumes that the polaron peak is a Lorentzian has the same weight
    no matter the eta. With these assumptions, we can determine the
    location and weight exactly using two points, each from a different
    eta calculation."""

    numerator1 = np.sqrt(eta * eta_prime)
    numerator2 = A * eta - Aprime * eta_prime
    den1 = Aprime * eta - A * eta_prime
    den2 = A * eta - Aprime * eta_prime
    loc = w - np.abs(numerator1 * numerator2 / np.sqrt(den1 * den2))
    area = np.pi * A * ((w - loc) ** 2 + eta**2) / eta
    return loc, area


def peak_location_and_weight_wstep(w, wprime, A, Aprime, eta):
    """Takes two points (w, A) and (wprime, Aprime) from the same eta
    calculation and assumes they lie on the same Lorentzian of the form
    f(x) = C (eta/pi) / ( eta**2 + (x-loc)**2 ). Solves the equation to fit
    a Lorentzian through those two points by determining C, x0."""

    firstterm = A * Aprime * (w - wprime) ** 2 / (A - Aprime) ** 2
    secondterm = eta**2
    loc1 = (w * A - wprime * Aprime) / (A - Aprime) + np.sqrt(
        firstterm - secondterm
    )
    loc2 = (w * A - wprime * Aprime) / (A - Aprime) - np.sqrt(
        firstterm - secondterm
    )
    area1 = np.pi * A * ((w - loc1) ** 2 + eta**2) / eta
    area2 = np.pi * A * ((w - loc2) ** 2 + eta**2) / eta
    if area1 > 1.0:
        return loc2, area2
    else:
        return loc1, area1


def peak_location_and_weight_scipy(wrange, Arange, eta):
    """Takes a bunch of points lying on the Lorentzian and does scipy.minimize
    fit of a Lorentzian function to it. Outputs fit parameters and error."""

    fitparams, error = curve_fit(
        lorentzian, wrange, Arange, p0=[wrange[-1], 1, eta]
    )

    return fitparams, error


def lorentzian(w, loc, scale, eta):
    return scale * (eta / np.pi) / ((w - loc) ** 2 + eta**2)



from contextlib import contextmanager
import sys
from warnings import warn

from loguru import logger


try:
    from mpi4py import MPI

    COMM = MPI.COMM_WORLD
    MPI_WORLD_SIZE = COMM.Get_size()
    MPI_RANK = COMM.Get_rank()

except ImportError:
    COMM = None
    MPI_WORLD_SIZE = 1
    MPI_RANK = 0


def generic_filter(names):
    if names == "all":
        return None

    def f(record):
        return record["level"].name in names

    return f


rank_string = str(MPI_RANK).zfill(3)
rank_string = f" ({rank_string}) "

DEBUG_FMT_WITH_MPI_RANK = (
    "<fg #808080>{time:YYYY-MM-DD HH:mm:ss.SSS} "
    "{name}:{function}:{line}</> "
    "|<lvl>{level: <10}</>|%s<lvl>{message}</>" % rank_string
)

DEBUG_FMT_WITHOUT_MPI_RANK = (
    "<fg #808080>{time:YYYY-MM-DD HH:mm:ss}</> <lvl>{message}</>"
)

WARN_FMT_WITHOUT_MPI_RANK = (
    "<fg #808080>{time:YYYY-MM-DD HH:mm:ss.SSS} "
    "{name}:{function}:{line}</> "
    "|<lvl>{level: <10}</>| <lvl>{message}</>"
)


def configure_loggers(
    stdout_filter=["INFO", "SUCCESS"],
    stdout_debug_fmt=DEBUG_FMT_WITH_MPI_RANK,
    stderr_filter_warnings=["WARNING"],
    stderr_filter=["ERROR", "CRITICAL"],
    stdout_fmt=DEBUG_FMT_WITHOUT_MPI_RANK,
    stderr_fmt=WARN_FMT_WITHOUT_MPI_RANK,
    enable_python_standard_warnings=False,
):
    """Configures the ``loguru`` loggers. Note that the loggers are initialized
    using the default values by default.

    .. important::

        ``logger.critical`` `always` terminates the program, either through
        ``COMM.MPI_Abort()`` if ``run_as_mpi`` is True, or ``sys.exit(1)``
        otherwise.

    Parameters
    ----------
    stdout_filter : list of str, optional
        List of logging levels to include in the standard output stream.
    stdout_debug_fmt : str, optional
        Loguru format for the special debug stream.
    stdout_fmt : str, optional
        Loguru format for the rest of the standard output stream.
    stderr_filter : list, optional
        List of logging levels to include in the standard error stream.
    stderr_fmt : str, optional
        Loguru format for the rest of the standard error stream.
    enable_python_standard_warnings : bool, optional
        Raises dummy warnings on ``logger.warning`` and ``logger.error``.
    """

    logger.remove(None)  # Remove ALL handlers

    # # Use the MPI ranks if available
    if "DEBUG" in stdout_filter:
        stdout_filter = [xx for xx in stdout_filter if xx != "DEBUG"]
        logger.add(
            sys.stdout,
            colorize=True,
            filter=generic_filter(["DEBUG"]),
            format=stdout_debug_fmt,
        )

    # Only log info, success and warnings on RANK == 0
    if MPI_RANK == 0:
        logger.add(
            sys.stdout,
            colorize=True,
            filter=generic_filter(stdout_filter),
            format=stdout_fmt,
        )
        logger.add(
            sys.stdout,
            colorize=True,
            filter=generic_filter(stderr_filter_warnings),
            format=stderr_fmt,
        )

    # Log errors and criticals on every rank
    logger.add(
        sys.stderr,
        colorize=True,
        filter=generic_filter(stderr_filter),
        format=stderr_fmt,
    )

    # We always exit on critical
    if MPI_WORLD_SIZE == 1:
        logger.add(lambda _: sys.exit(1), level="CRITICAL")
    else:
        logger.add(lambda _: COMM.Abort(), level="CRITICAL")

    if enable_python_standard_warnings:
        logger.add(lambda _: warn("DUMMY WARNING"), level="WARNING")
        logger.add(lambda _: warn("DUMMY ERROR"), level="ERROR")


def DEBUG():
    """Quick helper to enable DEBUG mode."""

    configure_loggers(stdout_filter=["DEBUG", "INFO", "SUCCESS"])


def _TESTING_MODE():
    """Enables a testing mode where loggers are configured as usual but where
    the logger.warning and logger.error calls actually also raise a dummy
    warning with the text "DUMMY WARNING" and "DUMMY ERROR", respectively.
    Used for unit tests."""

    configure_loggers(
        stdout_filter=["DEBUG", "INFO", "SUCCESS"],
        enable_python_standard_warnings=True,
    )


def DISABLE_DEBUG():
    """Quick helper to disable DEBUG mode."""

    configure_loggers(stdout_filter=["INFO", "SUCCESS"])


@contextmanager
def disable_logger():
    """Context manager for disabling the logger."""

    logger.disable("")
    try:
        yield None
    finally:
        logger.enable("")


@contextmanager
def _testing_mode():
    _TESTING_MODE()
    try:
        yield None
    finally:
        DEBUG()


@contextmanager
def debug():
    DEBUG()
    try:
        yield None
    finally:
        DISABLE_DEBUG()


DISABLE_DEBUG()



#!/usr/bin/env python3

# Run with e.g.
# mpiexec -np `n_processes` python 002_PETSc_simple_example.y
# to start in parallel

import numpy as np

# uncomment if want to visualize PETSc accuracy benchmark
# import matplotlib as mpl
# import matplotlib.pyplot as plt

from mpi4py import MPI

import sys
import os

script_dir = os.path.dirname(os.path.realpath(__file__))
head_ggce_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(head_ggce_dir)
    import ggce  # noqa: F401

# Import the parallel executor and load in the MPI communicator
from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import (
    ParallelSparseExecutorMUMPS,
)  # noqa: E402

COMM = MPI.COMM_WORLD

# load the "grouth truth" data from https://journals.aps.org/prb/
# abstract/10.1103/PhysRevB.82.085116, figure 5 center for comparison
literature_data = np.loadtxt(os.path.join(script_dir, "000_example_A.txt"))

# Select the input parameters and load in the parallel model
M, N = 3, 9
model = Model()
model.set_parameters(hopping=0.1)
model.add_coupling(
    "EdwardsFermionBoson", Omega=1.25, M=M, N=N, dimensionless_coupling=2.5
)

# Set up an executor to manage the calculation
executor = ParallelSparseExecutorMUMPS(model, "info", mpi_comm=COMM)
executor.prime()

# arrange a omega array and use list comprehension to execute calculations
wgrid = np.linspace(-5.5, -2.5, 100)

# Set rtol appropriately large (but still small enough) so that we don't get
# a ton of tolerance warnings
spectrum = executor.spectrum(0.5 * np.pi, wgrid, 0.005, rtol=1e-13)

# Results are returned on RANK 0 only
if COMM.Get_rank() == 0:
    spectrum = np.array([-s[0].imag / np.pi for s in spectrum])
    xx = np.array([wgrid.squeeze(), spectrum.squeeze()]).T
    np.savetxt(os.path.join(script_dir, "tmp_parallel_results.txt"), xx)

    # Compare with "ground truth" (see https://journals.aps.org/prb/
    # abstract/10.1103/PhysRevB.82.085116, figure 5 center) via:
    # fig, ax = plt.subplots(1, 1, figsize=(3, 2))
    # ax.plot(wgrid, spectrum / spectrum.max(), 'k', label = rf'PETSc ($M={M}, N={N}$)')
    # ax.plot(literature_data[:, 0], literature_data[:, 1] / literature_data[:,1].max(),\
    #                             'r--', linewidth=0.5, label='Ground truth')
    # ax.set_ylabel("$A(\pi/2, \omega)$ [normalized]")
    # ax.set_xlabel("$\omega$")
    # plt.legend(bbox_to_anchor=(1,1), loc="upper left")
    # plt.savefig(os.path.join(script_dir,f'petsc_mumps_vs_groundtruth_M_{M}_N_{N}.png'), format='png', bbox_inches='tight')



import numpy as np
import time
import os
import pickle
from pathlib import Path

from tqdm import tqdm

from petsc4py import PETSc

from ggce.logger import logger
from ggce.utils.physics import G0_k_omega
from ggce.utils.utils import chunk_jobs, padded_kw, float_to_list
from ggce.executors.solvers import Solver

BYTES_TO_GB = 1073741274


class MassSolver(Solver):
    """A base class to connect to PETSc's powerful parallel sparse solver
    tools, to calculate G(k,w) in parallel. This is an abstract base class
    built on top of the abstract Solver class. This base class has fundamental
    methods such as matrix construction. The solve methods, as well as
    convergence and memory tracking are implemented in the inherited classes,
    while some basic routines like greens_function() that are method-agnostic
    are implemented here."""

    @property
    def mpi_brigade(self):
        if self._brigade_size is not None:
            return int(self.mpi_rank / self._brigade_size)
        return 0

    @property
    def mpi_comm_brigadier(self):
        return self._mpi_comm_brigadier

    @property
    def brigade_size(self):
        if self._brigade_size is not None:
            return self._brigade_size
        return self._mpi_comm.Get_size()

    @property
    def brigades(self):
        if self._brigade_size is not None:
            return int(self._mpi_comm.Get_size() / self._brigade_size)
        return 1

    @property
    def brigade_rank(self):
        if self._brigade_size is not None:
            return self._mpi_comm_brigadier.Get_rank()
        return self.mpi_rank

    @property
    def matr_dir(self):
        """This property sets the directory where the method
        _scaffold_from_disk looks for pickled matrices (in CSR format).
        It is set in __init__
        """
        if self._matr_dir is None:
            logger.warning("matr_dir not set -- GGCE will construct matrices.")
        return self._matr_dir

    def __init__(self, brigade_size=None, matr_dir=None, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._matr_dir = matr_dir
        if matr_dir is not None:
            self._matr_dir = Path(matr_dir)
            self._matr_dir.mkdir(exist_ok=True, parents=True)

        if self._mpi_comm is None:
            logger.critical(
                "PETSc solver cannot run with "
                "mpi_comm=None. Pass MPI_COMM when "
                "instantiating the MassSolver."
            )

        # brigade split
        self._brigade_size = brigade_size
        if self._brigade_size is not None:
            self.split_into_brigades()
        else:
            logger.warning(
                "Only one brigade, no splitting required. "
                "Using original MPI_COMM."
            )
            self._mpi_comm_brigadier = self._mpi_comm

    def split_into_brigades(self):
        """Splits the MPI_COMM provided into 'brigades' of ranks operating
        together. Does this on the basis of the provided _brigade_size,
        by assigning a process to a brigade on the basis of its global
        rank modulo the brigade size. mpi_brigade is automatically
        evaluated as a @property of the class.

        If the ranks cannot be evenly divided into brigades, raises an error
        and terminates the calculation. In future releases brigade splits
        will be able to handle non-even division and/or adjust on the fly.

        Returns
        -------
        None
            New mpi_comm_brigadier are set as attributes of the class.
        """

        # for now the implementation has limitations: must have worldsize
        # evenly divided into brigades
        assert self.mpi_world_size % self._brigade_size == 0, logger.error(
            f"Number of MPI ranks {self.mpi_world_size} cannot be "
            f"equally divided into brigades with size {self._brigade_size}."
        )

        self._mpi_comm_brigadier = self._mpi_comm.Split(
            self.mpi_brigade, self.mpi_rank
        )

    def get_jobs_on_this_brigade(self, jobs):
        """Get's the jobs assigned to this group of ranks. Note this method
        silently behaves as it should when the world size is 1, and will log
        a warning if it is called but the communicator is not initialized.

        Parameters
        ----------
        jobs : list
            The jobs to chunk

        Returns
        -------
        list
            The jobs assigned to this brigade.
        """

        if self.brigades == 1:
            logger.warning("Chunking jobs with COMM_WORLD_SIZE=1")
            return jobs
        return chunk_jobs(jobs, self.brigades, self.mpi_brigade)

    def _setup_petsc_structs(self):
        """This function serves to initialize the various vectors and matrices
        (using PETSc data types) that are needed to solve the linear problem.
        They are set up using the sparse scheme, in parallel, so that each
        process owns only a small chunk of it.

        _mpi_comm_brigadier is used throughout to make sure that separate
        brigades work on separate (k,w) points, as is intended by the double-
        parallelization scheme.
        """

        # Initialize the parallel vector b from Ax = b
        self._vector_b = PETSc.Vec().create(comm=self._mpi_comm_brigadier)

        # Need to set the total size of the vector
        self._vector_b.setSizes(self._linsys_size)

        # This sets all the other PETSc options as defaults
        self._vector_b.setFromOptions()

        # Now we create the solution vector, x in Ax = b
        self._vector_x = self._vector_b.duplicate()

        # Now determine what is the local size PETSc picked
        _n_local = self._vector_b.getLocalSize()

        # Figure out what the given process owns
        self._rstart, self._rend = self._vector_b.getOwnershipRange()
        logger.debug(
            f"I am rank {self.mpi_rank} in brigade "
            f"{self.mpi_brigade} and got range "
            f"{self._rstart} to {self._rend}"
        )

        # Create the matrix for the linear problem
        self._mat_X = PETSc.Mat().create(comm=self._mpi_comm_brigadier)

        # set the matrix dimensions
        # input format is [(n,N),(m,M)] where capitals are total matrix
        # dimensions and lowercase are local block dimensions
        # see bottom of PETSc listserv entry
        # https://lists.mcs.anl.gov/mailman/htdig/petsc-users/2015-March/024879.html
        # for example
        self._mat_X.setSizes(
            [(_n_local, self._linsys_size), (_n_local, self._linsys_size)]
        )
        # This sets all the other PETSc options as defaults
        self._mat_X.setFromOptions()

        # This actually creates the matrix
        self._mat_X.setUp()

    def _sparse_matrix_from_equations(self, k, w, eta):
        """This function iterates through the GGCE equations dicts to extract
        the row, column coordiante and value of the nonzero entries in the
        matrix. This is subsequently used to construct the parallel sparse
        system matrix. This is exactly the same as in the Serial class: however
        that method returns X, v whereas here we need row_ind/col_ind_dat.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
        list, list, list
            The row and column coordinate lists, as well as a list of values of
            the matrix that are nonzero.
        """

        row_ind = []
        col_ind = []
        dat = []

        total_bosons = np.sum(self._system.model.phonon_number)
        for n_bosons in range(total_bosons + 1):
            for eq in self._system.equations[n_bosons]:
                row_dict = dict()
                index_term_id = eq.index_term.id()
                ii_basis = self._basis[index_term_id]

                for term in eq._terms_list + [eq.index_term]:
                    jj = self._basis[term.id()]
                    try:
                        row_dict[jj] += term.coefficient(k, w, eta)
                    except KeyError:
                        row_dict[jj] = term.coefficient(k, w, eta)

                row_ind.extend([ii_basis for _ in range(len(row_dict))])
                col_ind.extend([key for key, _ in row_dict.items()])
                dat.extend([value for _, value in row_dict.items()])

        # estimate sparse matrix memory usage
        # (complex (16 bytes) + int (4 bytes) + int) * nonzero entries
        est_mem_used = 24 * len(dat) / BYTES_TO_GB
        logger.debug(f"Estimated memory needed is {est_mem_used:.02f} MB")

        return row_ind, col_ind, dat

    def _scaffold(self, k, w, eta):
        """This function uses the GGCE equation sparse format data to construct
        a sparse matrix in the PETSc scheme.

        Parameters
        ----------
        k : float or array_like
            The momentum quantum number point of the calculation.
        w : float or array_like
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
        The matrices self._mat_X, self._vector_b are constructed in-place,
        nothing is returned.
        """

        self._linsys_size = len(self._basis)

        row_ind, col_ind, dat = self._sparse_matrix_from_equations(k, w, eta)

        # quickly report the sparsity of the matrix
        self._lengthdat = len(dat)
        self._sparsity = (
            self._linsys_size**2 - len(dat)
        ) / self._linsys_size**2
        self._edge_sparsity = len(dat) / self._linsys_size

        t0 = time.time()

        # parse out the nonzero (nnz) matrix structure across rows
        # so we can pre-allocate enough space for the matrix
        # avoid wasting space and speed up assembly ~ 20x

        # Call structs to initialize the PETSc vectors and matrices
        self._setup_petsc_structs()

        # set up arrays of length equal to space owned by a given MPI process
        # diag and offdiag store the number of nonzero entries in a given row
        # in the diagonal or off-diagonal block of the matrix
        diag_nnz = np.zeros(self._rend - self._rstart, dtype="i4")
        offdiag_nnz = np.zeros(self._rend - self._rstart, dtype="i4")

        # iterate through coo notation arrays to identify the nonzero entry
        # number in each row
        for i, elem in enumerate(row_ind):
            # check if this row / column is owned by this MPI process
            if self._rstart <= elem and elem < self._rend:
                if self._rstart <= col_ind[i] and col_ind[i] < self._rend:
                    diag_nnz[elem - self._rstart] += 1
                else:
                    offdiag_nnz[elem - self._rstart] += 1

        # pass the nnz arrays to PETSC matrix
        self._mat_X.setPreallocationNNZ((diag_nnz, offdiag_nnz))

        # now populate the matrix with actual values
        row_start = np.zeros(1, dtype="i4")
        col_pos = np.zeros(1, dtype="i4")
        val = np.zeros(1, dtype="complex128")
        for ii, row_coo in enumerate(row_ind):
            if self._rstart <= row_coo and row_coo < self._rend:
                row_start, col_pos, val = row_coo, col_ind[ii], dat[ii]
                self._mat_X.setValues(row_start, col_pos, val)

        # Assemble the matrix now that the values are filled in
        self._mat_X.assemblyBegin(self._mat_X.AssemblyType.FINAL)
        self._mat_X.assemblyEnd(self._mat_X.AssemblyType.FINAL)

        # Assign values for the b vector
        a = self._system.model.lattice_constant
        t = self._system.model.hopping
        G0 = G0_k_omega(k, w, a, eta, t)
        self._vector_b.setValues(self._linsys_size - 1, G0)

        # Need to assemble before use
        self._vector_b.assemblyBegin()
        self._vector_b.assemblyEnd()

        # TODO: check memory usage
        # presently not wrapped for Python

        dt = time.time() - t0
        logger.debug("PETSc matrix assembled", elapsed=dt)

    def _scaffold_from_disk(self, k, w, eta, matr_dir):
        """The function uses the GGCE equation sparse format data to construct
        a sparse matrix in the PETSc scheme. Instead of using the basis,
        it loads the CSR elements from disk. The passed parameters
        are used to load the correct file from disk.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.
        matr_dir : string (path)
            The absolute path of the location of the matrices to be loaded.

        Returns
        -------
        The matrices self._mat_X, self._vector_b are constructed in-place,
        nothing is returned.
        """

        # Get the total size of the linear system -- needed by PETSc
        assert matr_dir is not None
        self._linsys_size = self._get_matr_size(matr_dir)

        matrix_loc = os.path.join(
            matr_dir, f"matr_at_k_{k:.10f}_w_{w:.10f}_e_{eta:.10f}.pkl"
        )
        with open(matrix_loc, "rb") as datafile:
            row_ind, col_ind, dat = pickle.load(datafile)

        # quickly report the sparsity of the matrix
        self._lengthdat = len(dat)
        self._sparsity = (
            self._linsys_size**2 - len(dat)
        ) / self._linsys_size**2
        self._edge_sparsity = len(dat) / self._linsys_size
        t0 = time.time()

        # parse out the nonzero (nnz) matrix structure across rows
        # so we can pre-allocate enough space for the matrix
        # avoid wasting space and speed up assembly ~ 20x

        # Call structs to initialize the PETSc vectors and matrices
        self._setup_petsc_structs()

        # set up arrays of length equal to space owned by a given MPI process
        # diag and offdiag store the number of nonzero entries in a given row
        # in the diagonal or off-diagonal block of the matrix
        diag_nnz = np.zeros(self._rend - self._rstart, dtype="i4")
        offdiag_nnz = np.zeros(self._rend - self._rstart, dtype="i4")

        # iterate through coo notation arrays to identify the nonzero entry
        # number in each row
        for i, elem in enumerate(row_ind):
            # check if this row / column is owned by this MPI process
            if self._rstart <= elem and elem < self._rend:
                if self._rstart <= col_ind[i] and col_ind[i] < self._rend:
                    diag_nnz[elem - self._rstart] += 1
                else:
                    offdiag_nnz[elem - self._rstart] += 1

        # pass the nnz arrays to PETSC matrix
        self._mat_X.setPreallocationNNZ((diag_nnz, offdiag_nnz))

        # now populate the matrix with actual values
        row_start = np.zeros(1, dtype="i4")
        col_pos = np.zeros(1, dtype="i4")
        val = np.zeros(1, dtype="complex128")
        for ii, row_coo in enumerate(row_ind):
            if self._rstart <= row_coo and row_coo < self._rend:
                row_start, col_pos, val = row_coo, col_ind[ii], dat[ii]
                logger.debug(
                    f"I am rank {self.mpi_rank} and I am setting"
                    f" the values at {(row_start, col_pos)}"
                )
                self._mat_X.setValues(row_start, col_pos, val)

        # Assemble the matrix now that the values are filled in
        self._mat_X.assemblyBegin(self._mat_X.AssemblyType.FINAL)
        self._mat_X.assemblyEnd(self._mat_X.AssemblyType.FINAL)

        # Assign values for the b vector
        a = self._system.model.lattice_constant
        t = self._system.model.hopping
        G0 = G0_k_omega(k, w, a, eta, t)
        self._vector_b.setValues(self._linsys_size - 1, G0)

        # Need to assemble before use
        self._vector_b.assemblyBegin()
        self._vector_b.assemblyEnd()

        # TODO: check memory usage
        # presently not wrapped for Python

        dt = time.time() - t0
        logger.debug(
            f"PETSc matrix assembled, built from disk at: {self._matr_dir}",
            elapsed=dt,
        )

    def check_conv_manual(self, pc, rtol):
        """This helper function checks PETSC convergence manually, by computing
        the residual r = b - Ax directly, with the preconditioners applied,
        and comparing it to the rtol*||b||_2 convergence criterion.

        Parameters
        ----------
        pc           : PETSc_PC
            This is the preconditioner context from PETSc. In particular
            it allows us to manually compute the residual by applying
            the preconditioner to the residual vector because we take the
            norm. With left preconditioning, which is the default, residual
            norm is taken with preconditioner applied on the left.

        Returns
        -------
        The residual check is conducted in place, nothing is returned.
        """

        # compute the residual and apply the preconditioner
        _vector_res = self._vector_b.copy()
        pc.apply(self._vector_b - self._mat_X(self._vector_x), _vector_res)
        _vector_res_norm = _vector_res.norm(PETSc.NormType.NORM_2)
        # tolerance comparison is based on rtol * b magnitude, which also needs
        # to be preconditioned
        _vector_b_condt = self._vector_b.duplicate()
        pc.apply(self._vector_b, _vector_b_condt)
        _vector_b_norm = _vector_b_condt.norm(PETSc.NormType.NORM_2)

        # create variable measuring how much tolerance is met / exceeded
        # if positive, we are in trouble
        self.tol_excess = _vector_res_norm - rtol * _vector_b_norm

        # do a manual residual check on head node
        if self.mpi_rank == 0:
            if self.tol_excess > 0:
                logger.warning(
                    f"Solution failed residual relative tolerance check. "
                    "Solutions likely not fully converged: "
                    f"res_norm ({_vector_res_norm:.02e}) > "
                    f"rtol * b_norm ({rtol*_vector_b_norm:.02e})"
                )
            else:
                logger.debug("Solution passed manual residual check.")

    def greens_function(self, k, w, eta, return_meta=False, pbar=False):
        """Solves for the greens_function using the PETSc solver backend.
        Computation is massively parallel over k,w and for each matrix at a
        given (k,w) point.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float, ndarray
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.
        return_meta : bool
            If True, returns a tuple of the Green's function and the dictionary
            containing meta information. If False, returns just the Green's
            function (the default is False).

        Returns
        -------
        np.ndarray
            The resultant Green's function array of shape nk by nw.
        """

        k = float_to_list(k)
        w = float_to_list(w)

        # the jobs MUST be evenly divisible between brigades
        # we will force pad the arrays if this is not the case
        # and raise a warning
        try:
            assert len(k) * len(w) % self.brigades == 0
        except AssertionError:
            logger.warning(
                "Number of jobs (k,w points) is not evenly "
                "divisible between brigades. Padding initiated."
                " If you don't want this, change your k, w arrays."
            )
            k, w = padded_kw(k, w, self.brigades)

        # Generate a list of tuples for the (k, w) points to calculate.
        jobs = [(_k, _w) for _k in k for _w in w]

        # check if working from disk or computing matrices on the fly
        if self._matr_dir is not None:
            logger.info(
                "Matrices are being loaded from disk. GGCE will not "
                "re-compute them."
            )
        else:
            logger.info(
                "Matrices solved by the engine are being computed on the "
                "fly from the basis."
            )

        # Chunk the jobs appropriately. Each of these lists look like the jobs
        # list above.
        jobs_on_brigade = self.get_jobs_on_this_brigade(jobs)
        self._total_jobs_on_this_brigade = len(jobs_on_brigade)
        # Get the results on this rank.
        s = []
        for (_k, _w) in tqdm(jobs_on_brigade, disable=not pbar):
            s.append(self.solve(_k, _w, eta))

        # Gather the results from the brigade commanders to "the general"
        # (global rank 0)
        all_results = self._mpi_comm.gather(s, root=0)
        # create placeholder variables for final bcast of results
        res = None
        meta = None

        # need to get rid of duplicates, since each rank in a brigade sends
        # a copy of the results from the brigade
        if self.mpi_rank == 0:
            results = []
            if self.brigade_size > 1:
                for n in range(self.brigades):
                    results.append(all_results[int(n * self.brigade_size)])
            else:
                results = all_results

            results = [xx[ii] for xx in results for ii in range(len(xx))]

            # a copy of the results of the whole brigade
            s = [xx[0] for xx in results]
            meta = [xx[1] for xx in results]
            res = np.array(s)

            # remove nans from the padding
            res = res[~np.isnan(res)]

            # Ensure the returned array has the proper shape
            res = res.reshape(len(k), len(w))

        # Broadcast the final result to all ranks
        # memory intensive but intuitive
        res = self._mpi_comm.bcast(res, root=0)
        meta = self._mpi_comm.bcast(meta, root=0)

        if return_meta:
            return (res, meta)
        return res

    def prepare_system(self, k, w, eta):
        """Prepare the sparse-represented system to be solved by another
        executor.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
            Nothing is returned, the matrix is dumped to disk.
        """

        row_ind, col_ind, dat = self._sparse_matrix_from_equations(k, w, eta)
        xx = [row_ind, col_ind, dat]

        matr_loc = os.path.join(
            self.matr_dir, f"matr_at_k_{k:.10f}_w_{w:.10f}_e_{eta:.10f}.pkl"
        )
        with open(matr_loc, "wb") as matr_file:
            pickle.dump(xx, matr_file)

    def prepare_greens_function(
        self, k, w, eta, return_meta=False, pbar=False
    ):
        """Prepares matrices for the greens_function in parallel.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
            Nothing is returned -- the saved matrices are on disk.
        """

        k = float_to_list(k)
        w = float_to_list(w)

        # the jobs MUST be evenly divisible between brigades
        # we will force pad the arrays if this is not the case
        # and raise a warning
        try:
            assert len(k) * len(w) % self.brigades == 0
        except AssertionError:
            logger.warning(
                "Number of jobs (k,w points) is not evenly "
                "divisible between brigades. Padding initiated."
                " If you don't want this, change your k, w arrays."
            )
            k, w = padded_kw(k, w, self.brigades)

        # Generate a list of tuples for the (k, w) points to calculate.
        jobs = [(_k, _w) for _k in k for _w in w]

        # Chunk the jobs appropriately. Each of these lists look like the jobs
        # list above.
        jobs_on_brigade = self.get_jobs_on_this_brigade(jobs)
        self._total_jobs_on_this_brigade = len(jobs_on_brigade)

        logger.info(
            f"Running GGCE-PETSc in matrix prep mode. "
            f"Matrices are being saved to {self._matr_dir}."
        )

        # Get the results on this rank.
        for (_k, _w) in tqdm(jobs_on_brigade, disable=not pbar):
            self.prepare_system(_k, _w, eta)

        return

    @staticmethod
    def _k_omega_eta_to_str(k, omega, eta):
        # Note this will have to be redone when k is a vector in 2 and 3D!
        return f"{k:.10f}_{omega:.10f}_{eta:.10f}"

    @staticmethod
    def _get_matr_size(matr_dir):

        """For use with the _scaffold_from_disk method. Helps figure
        out the ultimate matrix size before loading all in."""

        all_files = os.listdir(matr_dir)
        all_files = [elem for elem in all_files if ".pkl" in elem]
        random_matr = np.random.choice(all_files)
        sample_matrix = os.path.join(matr_dir, random_matr)
        with open(sample_matrix, "rb") as datafile:
            row_ind, col_ind, dat = pickle.load(datafile)

        matrsize = max(row_ind) + 1

        return matrsize



import pytest

import numpy as np
import warnings

from ggce.logger import _testing_mode
from ggce.engine.terms import Config


class TestConfig:
    @staticmethod
    def test_MSONable():
        config = Config(np.array([[1, 2, 3, 4]]))
        d = config.as_dict()
        config2 = Config.from_dict(d)
        assert np.all(config.config == config2.config)

    @staticmethod
    def test_id_Green(
        Random1dPhononArray, Random2dPhononArray, Random3dPhononArray
    ):
        assert Config(Random1dPhononArray.copy() * 0).id() == "G"
        assert Config(Random2dPhononArray.copy() * 0).id() == "G"
        assert Config(Random3dPhononArray.copy() * 0).id() == "G"

    @staticmethod
    def test_1d_check_config_errorstate_dim_too_few():
        with pytest.raises(SystemExit):
            Config(np.array([0, 1, 2, 3, 4]))

    @staticmethod
    def test_1d_check_config_errorstate_negative_phonon_occupation(
        Random3dPhononArray,
    ):
        arr = Random3dPhononArray.copy()
        arr[2, 0, 1, 3] = -1
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_1d_check_config_warning():

        arr = (np.random.random(size=(2, 3, 3, 3, 4)) + 1.5).astype(int)
        with _testing_mode():
            with warnings.catch_warnings(record=True) as w:
                Config(arr)

        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "DUMMY WARNING" in str(w[-1].message)

    @staticmethod
    @pytest.mark.parametrize(
        "n_phonons",
        [1, 3, 5],
    )
    def test_n_phonon_types(n_phonons):
        arr = (np.random.random(size=(n_phonons, 3, 2, 3)) + 1.5).astype(int)
        config = Config(arr)
        assert config.n_phonon_types == n_phonons

    @staticmethod
    @pytest.mark.parametrize("dims", [1, 2, 3])
    def test_phonon_cloud_shapes(dims):
        arr = (np.random.random(size=([2] + [3] * dims)) + 1.5).astype(int)
        config = Config(arr)
        assert config.phonon_cloud_shape == tuple([3] * dims)

    @staticmethod
    @pytest.mark.parametrize("dims", [1, 2, 3])
    def test_total_phonons_per_type_and_total_phonons(dims):
        size = tuple([ii + 4 for ii in range(dims)])
        ptype1 = (np.random.random(size=size) + 1.5).astype(int)
        ptype2 = (np.random.random(size=size) + 2.5).astype(int)
        ptype3 = (np.random.random(size=size) + 3.5).astype(int)
        arr = np.array([ptype1, ptype2, ptype3])
        config = Config(arr)
        sarray = np.array([ptype1.sum(), ptype2.sum(), ptype3.sum()])
        assert np.all(config.total_phonons_per_type == sarray)
        assert config.total_phonons == sarray.sum()

    @staticmethod
    def test_Greens_config(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr *= 0
        Config(arr)

    @staticmethod
    def test_validate_critical_1d(Random1dPhononArray):
        arr = Random1dPhononArray.copy()
        arr[1, 2] = -1
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_2d(Random2dPhononArray):
        arr = Random2dPhononArray.copy()
        arr[1, 0, 1] = -1
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_3d(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr[1, 2, 1, 1] = -1
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_1d_left(Random1dPhononArray):
        arr = Random1dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, 0] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_1d_right(Random1dPhononArray):
        arr = Random1dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, -1] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_2d_left(Random2dPhononArray):
        arr = Random2dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, 0, :] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_2d_right(Random2dPhononArray):
        arr = Random2dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, -1, :] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_2d_up(Random2dPhononArray):
        arr = Random2dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, :, 0] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_2d_down(Random2dPhononArray):
        arr = Random2dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, :, -1] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_3d_left(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, 0, :, :] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_3d_right(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, -1, :, :] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_3d_up(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, :, 0, :] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_3d_down(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, :, -1, :] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_3d_front(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, :, :, 0] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_validate_critical_config_edges_3d_back(Random3dPhononArray):
        arr = Random3dPhononArray.copy()
        arr += 1
        Config(arr.copy())
        arr[:, :, :, -1] = 0
        with pytest.raises(SystemExit):
            Config(arr)

    @staticmethod
    def test_add_phonon_max_modifications(Random1dPhononArray):
        config = Config(Random1dPhononArray.copy())
        config.add_phonon_(0, 1)
        with pytest.raises(SystemExit):
            config.add_phonon_(1, 2)

    @staticmethod
    def test_remove_phonon_max_modifications(Random1dPhononArray):
        config = Config(Random1dPhononArray.copy() + 3)
        config.remove_phonon_(0, 1)
        with pytest.raises(SystemExit):
            config.remove_phonon_(1, 2)

    @staticmethod
    def test_add_remove_dimension_mismatch(Random3dPhononArray):
        config = Config(Random3dPhononArray.copy())
        with pytest.raises(SystemExit):
            config.add_phonon_(1, 0, 0, 0, 0)
        config = Config(Random3dPhononArray.copy())
        with pytest.raises(SystemExit):
            config.remove_phonon_(1, 0, 0, 0, 0)

    @staticmethod
    def test_remove_phonon_index_dne(Random3dPhononArray):
        config = Config(Random3dPhononArray.copy())
        with pytest.raises(SystemExit):
            config.remove_phonon_(1, 0, 0, 20)

    @staticmethod
    def test_remove_phonon_negative_occupancy_after_removal(
        Random3dPhononArray,
    ):
        arr = Random3dPhononArray.copy()
        arr[0, 2, 3, 3] = 0
        config = Config(arr)
        with pytest.raises(SystemExit):
            config.remove_phonon_(0, 2, 3, 3)

    @staticmethod
    def test_remove_phonon_shift_1d_right():
        arr = np.array([[1, 2, 3, 0, 0, 0, 1], [4, 5, 6, 0, 0, 0, 0]])
        config = Config(arr.copy())
        shift = config.remove_phonon_(0, 6)
        assert np.all(shift == np.array([0]))
        assert np.all(config.config == np.array([[1, 2, 3], [4, 5, 6]]))

    @staticmethod
    def test_remove_phonon_shift_1d_left():
        arr = np.array([[0, 0, 0, 0, 1, 2, 3], [1, 0, 0, 0, 4, 5, 6]])
        config = Config(arr.copy())
        shift = config.remove_phonon_(1, 0)
        assert np.all(shift == np.array([4]))
        assert np.all(config.config == np.array([[1, 2, 3], [4, 5, 6]]))

    @staticmethod
    def test_add_phonon_Shift_1d_right():
        arr = np.array([[1, 2, 3], [4, 5, 6]])
        config = Config(arr.copy())
        shift = config.add_phonon_(0, 6)
        arr2 = np.array([[1, 2, 3, 0, 0, 0, 1], [4, 5, 6, 0, 0, 0, 0]])
        assert np.all(shift == np.array([0]))
        assert np.all(config.config == arr2)

    @staticmethod
    def test_add_phonon_Shift_1d_left():
        arr = np.array([[1, 2, 3], [4, 5, 6]])
        config = Config(arr.copy())
        shift = config.add_phonon_(1, -4)
        arr2 = np.array([[0, 0, 0, 0, 1, 2, 3], [1, 0, 0, 0, 4, 5, 6]])
        assert np.all(shift == np.array([4]))
        assert np.all(config.config == arr2)

    @staticmethod
    def test_remove_phonon_shift_2d_0():
        arr = np.array([[[1, 2, 3, 0, 0, 0, 1], [4, 5, 6, 0, 0, 0, 0]]])
        config = Config(arr.copy())
        shift = config.remove_phonon_(0, 0, 6)
        assert np.all(shift == np.array([0, 0]))
        assert np.all(config.config == np.array([[[1, 2, 3], [4, 5, 6]]]))

    @staticmethod
    def test_remove_phonon_shift_2d_1():
        arr = np.array([[[0, 2, 3], [1, 0, 0]]])
        config = Config(arr.copy())
        shift = config.remove_phonon_(0, 1, 0)
        assert np.all(shift == np.array([0, 1]))
        assert np.all(config.config == np.array([[[2, 3]]]))



############ Stress-testing PETSc ##########
"""
In this script we push PETSc to the limit. We will calculate entire spectral
slices, tracking both time to evaluate as well as saving the actual spectra.
We are looking to evalaute convergence wiht (M,N), as well as time to calculate
"""
########### ################

# before we import numpy, set the number of threads for numpy multithreading
# using the OMP_NUM_THREADS environment variable
import numpy as np

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os
import time

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(ggce_head_dir)
    import ggce

# load the mpi communicator
from mpi4py import MPI

# load the Model and Executor classes to store parameters and manage the calculation
from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS

# run PETSc benchmark for cloud size up to cloud_max with up to bosons_max bosons
cloud_ext = 8
bosons_max = 10

Omega = 2.5
coupling = 1.0
hopping = 1.0
eta = 0.005

# define w and k grid for the calculation
wgrid = np.linspace(-4.0, -2.0, 5)
k = 0.0

# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

# start by creating a folder with the text and image output in run directory
# only head node make directories
results_dir = os.path.join(script_dir, "check_mem_eff")
spectra_vis_path = os.path.join(results_dir, "spectra_visuals")
text_data_path = os.path.join(results_dir, "text_data")

if COMM.Get_rank() == 0:
    if not os.path.exists(results_dir):
        os.mkdir(results_dir)
    # now create nested directories for text and image respectively
    if not os.path.exists(spectra_vis_path):
        os.mkdir(spectra_vis_path)
    if not os.path.exists(text_data_path):
        os.mkdir(text_data_path)

# initialize the model with pre-determined parameters
model = Model()
model.set_parameters(hopping=hopping)
model.add_coupling(
    "Holstein",
    Omega=Omega,
    M=cloud_ext,
    N=bosons_max,
    dimensionless_coupling=coupling,
)

# initialize the Executor to manage the calculation
executor = ParallelSparseExecutorMUMPS(model, "debug", mpi_comm=COMM)
executor.prime()

# execute the solve
results = [executor.solve(k, w, eta=eta) for w in wgrid]

# only collect and output results on the head node
if COMM.Get_rank() == 0:

    # process the solver output to extract time to solve and spectrum vals
    greens_func, exitcodes_dict = np.array(results).T
    spectrum = np.array([-s.imag / np.pi for s in greens_func])
    times = np.array([exitcodes["time"][0] for exitcodes in exitcodes_dict])
    mumps_conv_codes = np.array(
        [exitcodes["mumps_exit_code"][0] for exitcodes in exitcodes_dict]
    )
    mumps_mem_tot = np.array(
        [exitcodes["mumps_mem_tot"][0] for exitcodes in exitcodes_dict]
    )
    tol_excess = np.array(
        [
            exitcodes["manual_tolerance_excess"][0]
            for exitcodes in exitcodes_dict
        ]
    )

    fig, ax = plt.subplots(1, 1, figsize=(6, 4))
    ax.plot(
        wgrid,
        spectrum / spectrum.max(),
        "k",
        label=rf"PETSc ($M={cloud_ext}, N={bosons_max}$)",
    )
    ax.set_ylabel(f"$A(0, \omega)$ [normalized]")
    ax.set_xlabel("$\omega$")
    plt.legend(bbox_to_anchor=(1, 1), loc="upper left")
    plt.tight_layout()
    plt.savefig(
        os.path.join(
            spectra_vis_path, f"scope_M_{cloud_ext}_N_{bosons_max}.png"
        ),
        format="png",
    )

    # also output the spectrum data to disk for posterity / postprocess
    xx = np.array(
        [wgrid, spectrum, times, mumps_conv_codes, mumps_mem_tot, tol_excess]
    ).T
    np.savetxt(
        os.path.join(
            text_data_path,
            f"parallel_results_M_{cloud_ext}_N_{bosons_max}.txt",
        ),
        xx,
        header=f"omega                         spectral func               time_to_compute (sec)    "
        f"MUMPS convergence code (0 = good)  MUMPS total memory used (Gb)  "
        f"Tolerance excess  (|r| - rtol*|b|)",
        delimiter="    ",
    )



import cmath
import numpy as np

from monty.json import MSONable

from ggce.logger import logger
from ggce.utils import physics
from ggce.model import SingleTerm


def _config_shape_legal(config):
    if len(config.shape) < 2:
        return False
    return True


def _config_values_legal(config):
    if np.any(config < 0):
        return False
    return True


def _config_shape_small_dimension(config):
    if len(config.shape) > 4:
        return False
    return True


def _check_config(config):
    if not _config_shape_legal(config):
        logger.critical(f"Provided config {config} has <2 dimensions")
    if not _config_values_legal(config):
        logger.critical(f"Invalid config {config}: some < 0")
    if not _config_shape_small_dimension(config):
        logger.warning(
            f"Provided config of shape {config.shape} is greater than 4, "
            "indicating more than 3 spatial dimensions. This type of case "
            "is untested and the user should proceed with caution."
        )


def _config_edges_legal(config):

    # First, sum up over all of the phonon types. This produces the
    # "anchor" config.
    _config = config.sum(axis=0)

    # For every dimension, we need to check the "edge". This amounts to
    # swapping the axes and summing
    n_dimensions = len(_config.shape)  # The number of dimensions
    for ii in range(n_dimensions):
        edge_left = _config.swapaxes(0, ii)[0, ...]
        edge_right = _config.swapaxes(0, ii)[-1, ...]
        if np.sum(edge_left) == 0 or np.sum(edge_right) == 0:
            return False
    return True


def _extent_of_1d(config1d):
    """Gets the extent of a 1d vector. TODO: this will require a more general
    solution to find the extent of 2 and 3d systems."""

    where_nonzero = np.where(config1d != 0)[0]
    L = len(where_nonzero)
    if L < 2:
        return L
    minimum_index = min(where_nonzero)
    maximum_index = max(where_nonzero)
    return maximum_index - minimum_index + 1


def config_legal(
    config,
    max_phonons_per_site=None,
    phonon_extent=None,
    phonon_number=None,
    allow_green=False,
):
    """Helper method designed to test a standalone configuration of phonons
    for legality. A `legal` ``config`` array satisfies the following
    properties:

    - It has at least two axes (one phonon index and one spatial)
    - It has only nonnegative entries
    - Every spatial "edge" has at least one phonon of some type
    - The maximum number of phonons per site condition is satisfied
    - The phonon extent criterion is satisfied for all phonon types

    Parameters
    ----------
    config : numpy.array
        The input configuration.
    max_phonons_per_site : int, optional
        If not None, checks that the config satisfies the maximum number of
        phonons per site. Returns False if any site contains more than the
        specified number.
    phonon_extent : list, optional
        A list of int where each entry is the extent allowed for that phonon
        type. Checks this if not None, otherwise ignores.
    allow_green : bool, optional
        If True, automatically passes the check if the total number of phonons
        sum to 0.

    Returns
    -------
    bool
        Returns True if the config passes all the tests. False if it fails any
        of them.
    """

    if allow_green and np.sum(config) == 0:
        return True
    if not _config_shape_legal(config):
        return False
    if not _config_values_legal(config):
        return False
    if not _config_edges_legal(config):
        return False
    if max_phonons_per_site is not None:
        if not np.all(config <= max_phonons_per_site):
            return False
    if phonon_extent is not None:
        if any(
            [
                phonon_extent[ii] < _extent_of_1d(c1d)
                for ii, c1d in enumerate(config)
            ]
        ):
            return False
    if phonon_number is not None:
        if any(
            [
                np.sum(config[ii, ...]) > phonon_number[ii]
                for ii in range(config.shape[0])
            ]
        ):
            return False
    return True


def _validate_config_is_legal(config):
    """Helper function for exposing the :class:`Config.validate` method to
    other modules."""

    _check_config(config)

    if np.sum(config) == 0:
        return

    if not _config_edges_legal(config):
        logger.critical("Edge invalid on config axis")


class Config(MSONable):
    """A class for holding phonon occupations and defining operations on the
    cloud."""

    @property
    def config(self):
        """An array of the shape (n_phonon_types, cloud length axis 1, 2, ...).
        The array should only contain integers. Spatial information is
        contained in the indexes 1 and greater. The 0th index contains the
        information about the phonon type.

        Returns
        -------
        numpy.ndarray
        """

        return self._config

    @property
    def shape(self):
        """The shape of the config array.

        Returns
        -------
        tuple
        """

        return self.config.shape

    @property
    def n_spatial_dimensions(self):
        """The number of spatial dimensions in the config. Explicitly excludes
        counting the first dimension, which corresponds to the phonon index.

        Returns
        -------
        int
        """

        return len(self.shape) - 1

    @config.setter
    def config(self, x):
        _check_config(x)
        self._config = x

    @property
    def n_phonon_types(self):
        """The number of different types of phonons in the config. This is
        equal to the shape of the first axis of `config`.

        Returns
        -------
        int
        """

        return self._config.shape[0]

    @property
    def phonon_cloud_shape(self):
        """The shape corresponding to this particular cloud, ignoring
        information about phonon type.

        Returns
        -------
        tuple
        """

        return self._config.shape[1:]

    @property
    def total_phonons_per_type(self):
        """Gets the total number of phonons of each type. Essentially, sums
        over all axes except the first.

        Returns
        -------
        numpy.ndarray
        """

        axes = [ii + 1 for ii in range(len(self._config.shape) - 1)]
        return np.sum(self._config, axis=tuple(axes))

    @property
    def total_phonons(self):
        """Sums over the entire config. Returns the total number of phonons in
        the config.

        Returns
        -------
        int
        """

        return int(np.sum(self._config))

    def validate(self):
        """Checks the config for validity and throws a ``logger.critical``,
        which also kills the program if the validity checks fail. If any of the
        following fail, a critical is raised:

        - If any element in the config has less than 0 phonons.
        - If any of the edges of the cloud has 0 phonons total.

        .. warning::

            The Green's function, i.e. the case where there are actually 0
            phonons on the cloud, will not raise any critical.
        """

        _validate_config_is_legal(self._config)

    def __init__(self, config, max_modifications=1, modifications=0):
        """Initializes the Config class.

        Parameters
        ----------
        config : numpy.ndarray
            The configuration of phonons.
        max_modifications : int
            The maximum number of times one can modify the phonon config before
            throwing an error. This is precisely equal to the order of the
            phonon creation operators in V.
        modifications : int, optional
            The current number of modifications. Default is 0.
        """

        self._config = np.array(config).astype(int)
        self._max_modifications = max_modifications
        self._modifications = modifications
        self.validate()

    def __str__(self):
        if self.total_phonons == 0:
            return "G"
        rep = str(list(self._config.flatten()))
        shape = str(self._config.shape)
        return f"{rep}{shape}"

    def __repr__(self):
        return self.__str__()

    def id(self):
        """Returns the string representation of the config. This is constructed
        by flattening the config, converting to a list and then converting that
        list to a string. In order to ensure any possible ambiguity issues,
        the shape of the initial config is also used as an identifier.

        Returns
        -------
        str
        """

        return self.__str__()

    def _apply_phonon_reduction_rules_(self):
        """Executes the reduction rules. See remove_phonon_ for more
        details."""

        # No reduction necessary anymore, this is the Green's function
        if self.total_phonons == 0:
            return 0

        _config = self._config.sum(axis=0)

        shifts = []
        n_dimensions = len(_config.shape)  # The number of dimensions
        for ii in range(n_dimensions):
            _config = _config.swapaxes(0, ii)

            # Check where there are at least one phonon present at the 0th axis
            at_least_one_phonon_present = np.where(_config > 0)[0]

            # Find the bounds
            left = np.min(at_least_one_phonon_present).item()
            right = np.max(at_least_one_phonon_present).item()

            if right < _config.shape[0] - 1:
                _config = _config[: right + 1, ...]
                self._config = self._config.swapaxes(1, ii + 1)
                self._config = self._config[:, : right + 1, ...]
                self._config = self._config.swapaxes(1, ii + 1)
                shifts.append(0)

            elif left > 0:
                _config = _config[left:, ...]
                self._config = self._config.swapaxes(1, ii + 1)
                self._config = self._config[:, left:, ...]
                self._config = self._config.swapaxes(1, ii + 1)
                shifts.append(left)

            else:
                shifts.append(0)

            _config = _config.swapaxes(0, ii)

        return np.array(shifts)

    def remove_phonon_(self, *indexes):
        """Executes the removal of a phonon, and then all following phonon
        removal "reduction rules". I.e., removes slices of all zeros from the
        edges of the phonon clouds. Essentially Appendix A in this
        `PRB <https://journals.aps.org/prb/abstract/10.1103/
        PhysRevB.104.035106>`_. Specifically equations A1 and A3.

        Parameters
        ----------
        *indexes
            A tuple that must be of the same dimension as the config itself.
            The first value is the phonon index, and the others correspond to
            the real-space location in the phonon to remove. These indexes
            must fall within the existing config, else the program will abort.

        Returns
        -------
        numpy.ndarray
            An array of the index shifts used for modifying the coefficients
            of auxiliary Green's functions.
        """

        if self._modifications >= self._max_modifications:
            logger.critical(
                f"Max modifications {self._max_modifications} exceeded"
            )

        if len(indexes) != len(self._config.shape):
            logger.critical(
                f"Dimension mismatch between config and indexes {indexes}"
            )

        try:
            self._config[indexes] -= 1
        except IndexError:
            logger.critical(
                f"Index does not exist for {self._config} of shape "
                f"{self._config.shape} at {indexes}"
            )

        if self._config[indexes] < 0:
            logger.critical(
                f"Removal error: negative site occupancy for\n {self._config} "
                f"\nof shape {self._config.shape} at {indexes}"
            )

        # Actually modify the config object
        shift = self._apply_phonon_reduction_rules_()

        self._modifications += 1
        return shift

    def add_phonon_(self, *indexes):
        """Executes the addition of a phonon, and then all following phonon
        addition "reduction rules". Essentially Appendix A in this
        `PRB <https://journals.aps.org/prb/abstract/10.1103/
        PhysRevB.104.035106>`_. Specifically equations A2 and A4.

        Parameters
        ----------
        *indexes
            A tuple that must be of the same dimension as the config itself.
            The first value is the phonon index, and the others correspond to
            the real-space location in the phonon to remove. These indexes
            `can` fall outside of the current config, since phonons can be
            added anywhere on the chain.

        Returns
        -------
        numpy.ndarray
            An array of the index shifts used for modifying the coefficients
            of auxiliary Green's functions.
        """

        if self._modifications >= self._max_modifications:
            logger.critical(
                f"Max modifications {self._max_modifications} exceeded"
            )

        if len(indexes) != len(self._config.shape):
            logger.critical(
                f"Dimension mismatch between config and indexes {indexes}"
            )

        # Check that the phonon index is valid
        if indexes[0] < 0 or indexes[0] > self._config.shape[0] - 1:
            logger.critical(
                f"Phonon index {indexes[0]} invalid for config with "
                f"{self._config.shape[0]} unique phonon types"
            )

        spatial_indexes = np.array(indexes)[1:]
        zeros = np.array([0 for _ in range(len(spatial_indexes))])
        cloud_shape = np.array(self._config.shape)[1:]
        location_matrix = (zeros <= spatial_indexes) & (
            spatial_indexes < cloud_shape
        )

        # Easy case: the phonon type to add is in the existing cloud.
        # Here, no padding is required.
        if np.all(location_matrix):
            self._config[indexes] += 1
            self._modifications += 1
            return zeros  # Shift is all zeros in this case

        # Otherwise, we have to pad the array in various ways. First, it is
        # easy to calculate the shift. If an index is less than 0, that will
        # necessitate a shift. If an index is greater than the size of that
        # dimension, it will not necessitate a shift.
        shift = spatial_indexes.copy()
        shift[shift > 0] = 0

        # Now handle the padding. The first element is (0, 0) always, since we
        # never pad the phonon index part of the config
        pad = [(0, 0)]

        # Now, for each of the spatial dimensions we determine the padding
        pad = pad + [
            (0, 0)
            if 0 <= index < self._config.shape[ii + 1] - 1
            else (-index, 0)
            if index < 0
            else (0, index - self._config.shape[ii + 1] + 1)
            for ii, index in enumerate(spatial_indexes)
        ]

        # Update the config object accordingly
        self._config = np.pad(self._config, pad, "constant", constant_values=0)
        indexes = tuple([max(xx, 0) for xx in indexes])
        self._config[indexes] += 1

        self._modifications += 1
        return -shift


class Term(MSONable):
    """Base class for a single term in an equation. This object contains all
    information required for defining a term in the sum of equation 22 in the
    `PRB <https://journals.aps.org/prb/abstract/10.1103/
    PhysRevB.104.035106>`_. that this code is based on."""

    @property
    def config(self):
        """The configuration of phonons. The Config class contains the
        necessary methods for adding and removing phonons of various types
        and in various locations. See :class:`.Config`.

        Returns
        -------
        ggce.engine.terms.Config
        """

        return self._config

    @property
    def shape(self):
        """The shape of the config array.

        Returns
        -------
        tuple
        """

        return self.config.shape

    @config.setter
    def config(self, x):
        if not isinstance(x, Config):
            logger.critical(f"Invalid Config term of type {type(x)}")
        self._config = x

    @property
    def hamiltonian_term(self):
        """A single term in the coupling of the Hamiltonian.

        Returns
        -------
        ggce.model.SingleTerm
        """

        return self._hamiltonian_term

    @hamiltonian_term.setter
    def hamiltonian_term(self, x):
        if x is None:
            self._hamiltonian_term = None
            return
        if not isinstance(x, SingleTerm):
            logger.critical(f"Invalid Hamiltonian term of type {type(x)}")
        self._hamiltonian_term = x

    @property
    def constant_prefactor(self):
        """A constant prefactor that multiplies the evaluated value of the
        entire term when constructing the eventual matrix to invert at the end
        of the software pipeline.

        Returns
        -------
        complex
        """

        return self._constant_prefactor

    @constant_prefactor.setter
    def constant_prefactor(self, x):
        if not isinstance(x, (int, float)):
            logger.critical(f"Invalid constant prefactor {x}")
        self._constant_prefactor = x

    @property
    def exp_shift(self):
        """Factor which multiplies the entire final prefactor term in addition
        to :class:`Term.constant_prefactor`. It is given by

        .. math::

            e^{i \\mathbf{k} \\cdot \\boldsymbol{\\delta}
            \\odot \\mathbf{a}}

        where :math:`\\delta` is this ``exp_shift``. Furthermore, this term is
        determined by the relations between the different f-functions. It is
        modified during the reduction rules as noted in
        :class:`Config.remove_phonon_` and :class:`Config.add_phonon_`.

        .. warning::

            The shape of the ``exp_shift`` should be equal to the number of
            spatial dimensions of the phonon configuration.

        Returns
        -------
        numpy.ndarray
        """

        return self._exp_shift

    @exp_shift.setter
    def exp_shift(self, x):
        if x is None:
            self._exp_shift = None
            return
        if len(x) != self.config.n_spatial_dimensions:
            logger.critical(
                f"exp_shift {x} != config spatial dimension "
                f"{self.config.n_spatial_dimensions}"
            )
        self._exp_shift = x

    @property
    def f_arg(self):
        """This array corresponds to the value of the argument of the auxiliary
        Green's function, :math:`f_n(x)`, where :math:`x` is the ``f_arg``.
        If None, this implies that the ``f_arg`` value is left general. This
        should occur if a derived class is ``Index``-like, meaning it indexes
        an entire equation. This is a term that originally appears on the
        "left hand side" of the equation of motion. During generation of the
        specific equations, this will be set to a "true delta" value and that
        equation appended to the master list.

        Returns
        -------
        numpy.ndarray
        """

        return self._f_arg

    @f_arg.setter
    def f_arg(self, x):
        if x is None:
            self._f_arg = None
            return
        if len(x) != self.config.n_spatial_dimensions:
            logger.critical(
                f"f_arg {x} != config spatial dimension "
                f"{self.config.n_spatial_dimensions}"
            )
        self._f_arg = x

    @property
    def g_arg(self):
        """Identical to :class:`Term.f_arg`, but for the argument of the
        lattice Green's function :math:`g_0(x)`, where :math:`x` is ``g_arg``.

        Returns
        -------
        numpy.ndarray
        """

        return self._g_arg

    @g_arg.setter
    def g_arg(self, x):
        if x is None:
            self._g_arg = None
            return
        if len(x) != self.config.n_spatial_dimensions:
            logger.critical(
                f"g_arg {x} != config spatial dimension "
                f"{self.config.n_spatial_dimensions}"
            )
        self._g_arg = x

    def __str__(self):
        return str(self.config)

    def __repr__(self):
        return self.__str__()

    def __init__(
        self,
        config,
        hamiltonian_term=None,
        constant_prefactor=1.0,
        exp_shift=None,
        f_arg=None,
        g_arg=None,
    ):
        self._config = Config(config)
        self._hamiltonian_term = hamiltonian_term  # None is the index term
        self._constant_prefactor = constant_prefactor

        # Determines the argument of f and prefactors
        n_dims = self.config.n_spatial_dimensions
        if exp_shift is None:
            self._exp_shift = np.array([0.0 for _ in range(n_dims)])
        else:
            self._exp_shift = exp_shift  # exp(i*k*a*exp_shift)
        self._f_arg = f_arg  # Only None for the index term
        self._g_arg = g_arg  # => g(...) = 1

    def _get_phonon_config_id(self):
        """Returns a string of the phonon config id."""

        return "{" + self.config.id() + "}"

    def _get_f_arg_id(self):
        """Returns a string of the f_arg id."""

        if self._f_arg is None:
            return "(!)"

        return str(list(self._f_arg.astype(float)))

    def _get_g_arg_id(self):
        """Returns a string of the f_arg id."""

        if self._g_arg is None:
            return "<!>"

        return str(list(self._g_arg.astype(float)))

    def _get_c_exp_id(self):
        """Returns a string of the current value of the exponential shift."""

        if self._exp_shift is None:
            return "[!]"

        return str(list(self._exp_shift.astype(float)))

    def id(self, full=False):
        """Returns a string with which one can index the term. There are two
        types of identifiers the user can request. First is the full=False
        version, which produces a string of the form ``X(f_arg)``, where
        ``X`` is the configuration and ``f_arg`` is the argument of the
        auxiliary Green's function. Second is full=True, which returns the same
        string as full=False, but with ``g_arg`` and ``c_exp`` appended to the
        end as well. ``g_arg`` is the argument of :math:`g_0` and ``c_exp`` is
        the exponential shift indexes.

        Parameters
        ----------
        full : bool
            If true, returns also the g and exp shift terms with the
            id. This is purely for visualization and not for e.g.
            running any calculations, since those terms are pooled into the
            coefficient.

        Returns
        -------
        str
        """

        t1 = self._get_phonon_config_id() + self._get_f_arg_id()
        if not full:
            return t1
        t2 = self._get_g_arg_id() + self._get_c_exp_id()
        return t1 + t2

    def update_phonon_config_(self):
        """Specific update scheme needs to be run depending on whether we add
        or remove a phonon."""

        raise NotImplementedError

    def coefficient(self):
        raise NotImplementedError

    def _modify_n_phonons_(self):
        """By default, does nothing, will only be defined for the terms
        in which we add or subtract a phonon."""

        return

    def _increment_g_arg_(self, add_to_g_arg):
        """Increments the ``g_arg`` object by the provided value.

        Parameters
        ----------
        add_to_g_arg : numpy.ndarray
            The array to add to the ``g_arg`` attribute. If the provided value
            is not of the same shape as the current ``g_arg``, an error will be
            logged, since while this will not necessarily lead to the
            termination of the program, it is likely unintended.
        """

        if add_to_g_arg.shape != self._g_arg.shape:
            logger.error(f"Shape of {add_to_g_arg} != shape of {self._g_arg}")
        self._g_arg += add_to_g_arg

    def _set_f_arg_(self, val):
        raise NotImplementedError

    def check_if_green_and_simplify_(self):
        """There are cases when the algorithm produces results that look like
        e.g. ``G(1)[0]``. This corresponds to ``f_0(1)``, which is actually
        just a Green's function times a phase factor. The precise relation is

        .. math::

            f_0(\\boldsymbol{\\delta}) =
            e^{i \\mathbf{k} \\cdot \\boldsymbol{\\delta}
            \\odot \\mathbf{a}} G(k, w)

        We need to simplify this back to a term like G(0)[0], as if we don't,
        when the specific equations are generated, it will think there are
        multiple Greens equations, of which of course there can be only one."""

        if self._get_phonon_config_id() == "{G}" and np.any(self._f_arg != 0):
            self._exp_shift += self._f_arg
            self._f_arg = np.zeros_like(self._f_arg)


class IndexTerm(Term):
    """A term that corresponds to an index of the equation. Critical
    differences between this and the base class include that the ``f_arg``
    object is set to None by default and the constant_prefactor is set to 1.
    """

    def __init__(self, config):
        super().__init__(
            config=config,
            hamiltonian_term=None,
            constant_prefactor=1.0,
            exp_shift=None,
            f_arg=None,
            g_arg=None,
        )

    def _set_f_arg_(self, f_arg):
        """Overrides the set value of f_arg."""

        if self._hamiltonian_term is not None:
            logger.critical("Cannot set f_arg in an IndexTerm")
        self.f_arg = f_arg

    def _increment_g_arg_(self):
        raise NotImplementedError

    def coefficient(self, k, w, eta):
        """Coefficient of the IndexTerm (which is always 1).

        Parameters
        ----------
        k : float
        w : complex
        eta : float, optional

        Returns
        -------
        float
            The ``coefficient`` method in :class:`.IndexTerm` always returns 1.
        """

        return 1.0


class EOMTerm(Term):
    """A special instance of the base class that is part of the EOM of the
    Green's function."""

    def __init__(self, config, hamiltonian_term, model):
        super().__init__(config, hamiltonian_term)
        if self.hamiltonian_term is None:
            logger.critical(
                "EOM term requires a Hamiltonian term passed in the "
                "constructor"
            )
        self._model = model
        self._exp_shift = (
            self.hamiltonian_term.psi - self.hamiltonian_term.phi
        ).copy()
        self._f_arg = self.hamiltonian_term.phi.copy()
        self._constant_prefactor = self.hamiltonian_term.coupling
        self._lattice_constant = model.lattice_constant
        self._hopping = model.hopping

    def coefficient(self, k, w, eta):
        """This will set the prefactor to G0, since we assume
        that as a base class it originates from the g0 terms in the main
        EOM. Note that an index term does not have this method defined, since
        that terms prefactor should always be 1 (set by default to the
        constant prefactor), and this method will be overridden by
        AnnihilationTerm and CreationTerm classes."""

        exp_arg = 1j * np.dot(k, self._lattice_constant * self.exp_shift)

        return (
            self._constant_prefactor
            * physics.G0_k_omega(
                k, w, self._lattice_constant, eta, self._hopping
            )
            * cmath.exp(exp_arg)
        )

    def _increment_g_arg_(self, delta):
        return


class NonIndexTerm(Term):
    def __init__(self, config, hamiltonian_term, model, constant_prefactor):
        super().__init__(config, hamiltonian_term)
        assert self.hamiltonian_term is not None

        # This is entirely general now
        self.f_arg = self.hamiltonian_term.phi.copy()
        self.g_arg = (
            self.hamiltonian_term.psi - self.hamiltonian_term.phi
        ).copy()
        self.constant_prefactor = constant_prefactor
        self._model = model
        self.freq_shift = sum(
            [
                model.hamiltonian.phonon_frequencies[ii] * bpt
                for ii, bpt in enumerate(self.config.total_phonons_per_type)
            ]
        )
        self._lattice_constant = model.lattice_constant
        self._hopping = model.hopping

    def step_(self, *location):
        """Increments or decrements the phonons on the chain depending on the
        class derived class type."""

        self.g_arg += np.array(location)
        self.f_arg -= np.array(location)
        loc = [self.hamiltonian_term.phonon_index, *location]
        self._modify_n_phonons_(*loc)

    def coefficient(self, k, w, eta):

        exp_arg = 1j * np.dot(k, self._lattice_constant * self.exp_shift)
        exp_term = cmath.exp(exp_arg)

        w_freq_shift = w - self.freq_shift

        g_contrib = physics.g0_delta_omega(
            self.g_arg,
            w_freq_shift,
            self._lattice_constant,
            eta,
            self._hopping,
        )

        return self.constant_prefactor * exp_term * g_contrib


class AnnihilationTerm(NonIndexTerm):
    """Specific object for containing f-terms in which we must subtract
    a phonon from a specified site."""

    def _modify_n_phonons_(self, *loc):
        """Removes a phonon from the specified site corresponding to the
        correct type. Note this step is independent of the reductions of the
        f-functions to other f-functions."""

        shift = self.config.remove_phonon_(*loc)
        self.exp_shift -= shift
        self.f_arg += shift


class CreationTerm(NonIndexTerm):
    """Specific object for containing f-terms in which we must subtract
    a phonon from a specified site."""

    def _modify_n_phonons_(self, *loc):
        """This is done for the user in update_phonon_config_. Handles
        the phonon creation cases, since these can be a bit
        confusing. Basically, we have the possibility of creating a phonon on
        a site that is outside of the range of self.config. So we need methods
        of handling the case when an IndexError would've otherwise been raised
        during an attempt to increment an index that isn't there."""

        shift = self.config.add_phonon_(*loc)
        self.exp_shift = shift
        self.f_arg -= shift
